from functools import lru_cache
from typing import Any, Dict, List, Sequence, Tuple, Type, Union, Optional

import networkx as nx
from grai_client.schemas.edge import EdgeTypes
from grai_client.schemas.node import NodeTypes
from grai_client.schemas.schema import GraiType, Schema
from multimethod import multimethod


class GraphManifest:
    def __init__(self, nodes: Sequence[NodeTypes], edges: Sequence[EdgeTypes]):
        self.nodes: Sequence[NodeTypes] = nodes
        self.edges: Sequence[EdgeTypes] = edges

        self.node_index: Dict[str, Dict[str, NodeTypes]] = {}
        for node in nodes:
            self.node_index.setdefault(node.spec.namespace, {})
            self.node_index[node.spec.namespace].setdefault(node.spec.name, node)
        # assert all(isinstance(node.spec.id, uuid.UUID) for node in self.nodes), \
        #     "Graph manifests require node UUID's rather than name/namespace specification"
        # assert all(isinstance(edge.spec.source, uuid.UUID) and isinstance(edge.spec.destination, uuid.UUID)
        #            for edge in self.edges), \
        #     "Graph manifests require edge source/destination UUID's rather than name/namespace specification"

    def get_node(self, namespace: str, name: str) -> NodeTypes:
        return self.node_index[namespace][name]


class Graph:
    _container_key = 'obj'

    def __init__(self, manifest):
        self.manifest: GraphManifest = manifest
        self.graph = nx.DiGraph()
        self.graph.add_nodes_from(self.add_nodes_from_manifest())
        self.graph.add_edges_from(self.add_edges_from_manifest())

    def add_nodes_from_manifest(self):
        return ((hash(node.spec), {self._container_key: node})
                for node in self.manifest.nodes)

    def add_edges_from_manifest(self):
        return ((hash(edge.spec.source), hash(edge.spec.destination), {self._container_key: edge})
                for edge in self.manifest.edges)

    @lru_cache
    def get_node_id(self, namespace: str, name: str) -> int:
        node = self.manifest.get_node(namespace, name)
        return hash(node.spec)

    def get_node(self,
                 namespace: Optional[str] = None,
                 name: Optional[str] = None,
                 node_id: Optional[int] = None) -> GraiType:
        if namespace and name:
            node_id = self.get_node_id(namespace, name)
        elif not node_id:
            raise Exception(f"`get_node` requires either name & namespace or node_id argument")
        return self.graph.nodes.get(node_id)[self._container_key]

    def label(self, namespace: str, name: str) -> str:
        return self.get_node(namespace, name).spec.display_name

    def id_label(self, node_id: int) -> str:
        return self.get_node(node_id=node_id).spec.display_name

    def relabeled_graph(self):
        label_map = {hash(node.spec): f"{node.spec.namespace}-{node.spec.name}" for node in self.manifest.nodes}
        nodes = label_map.values()
        edges = ((hash(edge.spec.source), hash(edge.spec.destination), {self._container_key: edge})
                for edge in self.manifest.edges)
        edges = ((label_map[edge[0]], label_map[edge[1]], {}) for edge in edges)
        graph = nx.DiGraph()
        graph.add_nodes_from(nodes)
        graph.add_edges_from(edges)
        return graph


@multimethod
def process_items(vals: Any, version: Any, type: Any) -> List[GraiType]:
    message = (f"Process items does not have an implementation for "
               f"{vals=}, {version=}, {type=}")
    raise NotImplementedError(message)


@process_items.register
def process_dict(dict_item: Dict, version: str, type: str) -> GraiType:
    return Schema.to_model(dict_item, version, type)


@process_items.register
def process_node(item: NodeTypes, version: str, type: str) -> GraiType:
    return item


@process_items.register
def process_edge(item: EdgeTypes, version: str, type: str) -> GraiType:
    return item


@process_items.register
def process_sequence(item_iter: Sequence, version: str, type: str) -> List[GraiType]:
    return [process_items(item, version, type) for item in item_iter]


def build_graph(nodes: List[Dict], edges: List[Dict], version: str) -> Graph:
    nodes = process_items(nodes, version, 'Node')
    edges = process_items(edges, version, 'Edge')
    manifest = GraphManifest(nodes, edges)
    return Graph(manifest)



from grai_client.endpoints import list_clients
from grai_client.endpoints.client import BaseClient

from grai_graph.graph import build_graph


def graph_builder(self: BaseClient):
    nodes = self.get('nodes')
    edges = self.get('edges')
    return build_graph(nodes, edges, self.id)


for client in list_clients():
    client.build_graph = graph_builder



import uuid
from typing import Union

import networkx as nx

from grai_graph.graph import Graph


class GraphAnalyzer:
    def __init__(self, graph: Graph):
        self.graph = graph

    def downstream_nodes(self, namespace: str, name: str):
        node_id = self.graph.get_node_id(namespace, name)
        downstream = list(nx.bfs_successors(self.graph.graph, node_id))[0][1]
        return [self.graph.get_node(node_id=node) for node in downstream]

    def test_delete_node(self, namespace: str, name: str):
        return list(self.downstream_nodes(namespace, name))

    def test_type_change(self, namespace: str, name: str, new_type: str):
        node_id = self.graph.get_node_id(namespace, name)
        current_node = self.graph.get_node(namespace, name)
        if 'data_type' in current_node.spec.metadata:
            current_type = current_node.spec.metadata['data_type']
            if current_type == new_type:
                return []
        else:
            raise AttributeError(f"No data type defined for {self.graph.id_label(node_id)}")
        result = self.graph.graph.successors(node_id)
        return [self.graph.get_node(node_id=node_id) for node_id in result]





# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['grai_graph']

package_data = \
{'': ['*']}

install_requires = \
['grai-client>=0.1.3,<0.2.0',
 'networkx>=2.8.5,<3.0.0',
 'pydantic>=1.9.1,<2.0.0']

extras_require = \
{'vis': ['matplotlib>=3.5.2,<4.0.0', 'pydot>=1.4.2,<2.0.0']}

setup_kwargs = {
    'name': 'grai-graph',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Ian Eaves',
    'author_email': 'ian@grai.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
